﻿#synthDrivers/multilang-cjk.py
#A part of NonVisual Desktop Access (NVDA)
#This file is covered by the GNU General Public License.
#See the file COPYING for more details.
#Copyright 2011 World Light Information Limited and Hong Kong Blind Union.

from collections import OrderedDict
import globalVars
import speech
from logHandler import log
from characterProcessing import processSpeechSymbol
from languageHandler import getLanguage
from synthDriverHandler import SynthDriver,SynthSetting,NumericSynthSetting,VoiceInfo,getSynthList
from virtualSynthDriver import VirtualSynthDriver

_defaultCjkSynth="espeak"
_defaultWesternSynth="sapi5"

class SynthDriver(VirtualSynthDriver):
	name="multilang-cjk"
	description=_("CJK dual-voice virtual synthesizer")

	supportedSettings=(
		SynthSetting("synth1",_("Primar&y synthesizer")),
		NumericSynthSetting("rate1",_("Primary &rate")),
		NumericSynthSetting("pitch1",_("Primary &pitch")),
		NumericSynthSetting("inflection1",_("Primary &inflection")),
		NumericSynthSetting("volume1",_("Primary v&olume")),
		SynthSetting("synth2",_("&English synthesizer")),
		NumericSynthSetting("rate2",_("English r&ate")),
		NumericSynthSetting("pitch2",_("English pi&tch")),
		NumericSynthSetting("inflection2",_("English in&flection")),
		NumericSynthSetting("volume2",_("English vol&ume")),
		SynthSetting("cjk_number",_("Read nu&mbers using")),
		SynthSetting("spell_digit",_("Speak individual &digit")),
	)

	cjkSynth=_defaultCjkSynth
	westernSynth=_defaultWesternSynth
	cjkNumber=True
	spellDigit=True

	@classmethod
	def check(cls):
		return True

	def __init__(self):
		# Setup default parameters
		self.synth1=_defaultCjkSynth
		self.rate1=50
		self.pitch1=50
		self.inflection1=50
		self.volume1=100
		self.synth2=_defaultWesternSynth
		self.rate2=50
		self.pitch2=50
		self.inflection2=50
		self.volume2=100

		# Build voice dictionary from available synth drivers
		self.synthDict=OrderedDict()
		for x in getSynthList():
			if x[0]==self.name: continue
			self.synthDict[x[0]]=VoiceInfo(unicode(x[0],"utf-8"),x[1])

	def initSynths(self):
		self.unloadAllSynthsExcept((self.cjkSynth,self.westernSynth))
		if not self.loadSynth(self.cjkSynth):
			# Fall back to espeak upon failure
			self.cjkSynth="espeak"
		if not self.loadSynth(self.westernSynth):
			# Fall back to espeak upon failure
			self.westernSynth="espeak"
		self.setMasterSynth(self.cjkSynth)

	def _get_availableSynth1s(self):
		# CJK synthesizer
		return self.synthDict

	def _get_synth1(self):
		return self.cjkSynth

	def _set_synth1(self,synth1):
		self.cjkSynth=synth1
		self.initSynths()

	def _get_rate1(self):
		return self.getSynthSetting(self.cjkSynth, "rate")

	def _set_rate1(self,rate):
		return self.setSynthSetting(self.cjkSynth, "rate", rate)

	def _get_pitch1(self):
		return self.getSynthSetting(self.cjkSynth, "pitch")

	def _set_pitch1(self,pitch):
		return self.setSynthSetting(self.cjkSynth, "pitch", pitch)

	def _get_inflection1(self):
		return self.getSynthSetting(self.cjkSynth, "inflection")

	def _set_inflection1(self,inflection):
		return self.setSynthSetting(self.cjkSynth, "inflection", inflection)

	def _get_volume1(self):
		return self.getSynthSetting(self.cjkSynth, "volume")

	def _set_volume1(self,volume):
		return self.setSynthSetting(self.cjkSynth, "volume", volume)

	def _get_availableSynth2s(self):
		# Western synthesizer
		return self.synthDict

	def _get_synth2(self):
		return self.westernSynth

	def _set_synth2(self,synth2):
		self.westernSynth=synth2
		self.initSynths()

	def _get_synth2(self):
		return self.westernSynth

	def _get_rate2(self):
		return self.getSynthSetting(self.westernSynth, "rate")

	def _set_rate2(self,rate):
		return self.setSynthSetting(self.westernSynth, "rate", rate)

	def _get_pitch2(self):
		return self.getSynthSetting(self.westernSynth, "pitch")

	def _set_pitch2(self,pitch):
		return self.setSynthSetting(self.westernSynth, "pitch", pitch)

	def _get_inflection2(self):
		return self.getSynthSetting(self.westernSynth, "inflection")

	def _set_inflection2(self,inflection):
		return self.setSynthSetting(self.westernSynth, "inflection", inflection)

	def _get_volume2(self):
		return self.getSynthSetting(self.westernSynth, "volume")

	def _set_volume2(self,volume):
		return self.setSynthSetting(self.westernSynth, "volume", volume)

	def _get_availableSpell_digits(self):
		dict=OrderedDict()
		dict[u"yes"]=VoiceInfo(u"yes",_("Yes"))
		dict[u"no"]=VoiceInfo(u"no",_("No"))
		return dict

	def _get_spell_digit(self):
		return "yes" if self.spellDigit else "no"

	def _set_spell_digit(self,spell_digit):
		self.spellDigit=True if spell_digit=="yes" else False

	def _get_availableCjk_numbers(self):
		dict=OrderedDict()
		dict[u"primary"]=VoiceInfo(u"primary",_("Primary synthesizer"))
		dict[u"english"]=VoiceInfo(u"english",_("English synthesizer"))
		return dict

	def _get_cjk_number(self):
		return "primary" if self.cjkNumber else "english"

	def _set_cjk_number(self,cjk_number):
		self.cjkNumber=True if cjk_number=="primary" else False

	@classmethod
	def isWestern(cls,c):
		if (c>='A' and c<='Z') or (c>='a' and c<='z'):
			return True
		return False

	def speakNow(self,text,index):
		locale=getLanguage()
		prevSynth=self.cjkSynth
		portion=u""
		maxLen=len(text)
		for i in range(maxLen+1):
			if i==maxLen:
				# Always speak last portion at end of text
				c=u""
				synth=None
			else:
				c=text[i]
				if c<=' ':
					# Space and control codes do not alter language
					synth=prevSynth
				elif c>='0' and c<='9':
					# CJK synth read digits
					if self.spellDigit: c+=' '
					synth=self.cjkSynth if self.cjkNumber else self.westernSynth
				elif c=='.':
					# Check if full-stop is between digits
					if i==0 or i>=(maxLen-1):
						synth=self.westernSynth
					elif text[i-1]<'0' or text[i-1]>'9':
						synth=self.westernSynth
					elif text[i+1]<'0' or text[i+1]>'9':
						synth=self.westernSynth
					elif not self.cjkNumber:
						synth=self.westernSynth
					else:
						# CJK synth read decimal point
						c=processSpeechSymbol(locale,u'decimal point')
						if len(c)==0:
							c=processSpeechSymbol(locale,u'.')
						synth=self.cjkSynth
				elif self.isWestern(c):
					# Western synth read western characters
					synth=self.westernSynth
				elif ord(c)<=255 and i>0 and self.isWestern(text[i-1]):
					# Western synth read symbol after western text
					synth=self.westernSynth
				elif ord(c)<=255 and i<(maxLen-1) and self.isWestern(text[i+1]):
					# Western synth read symbol before western text
					synth=self.westernSynth
				else:
					# CJK synth read everything else
					synth=self.cjkSynth
			if synth==prevSynth:
				portion+=c
			else:
				portion=portion.strip()
				if len(portion)>0:
					self.speakPortion(prevSynth,portion,index)
				prevSynth=synth
				portion=c

	def speak(self,speechSequence):
		lastIndex=None
		charMode=False
		for item in speechSequence:
			if isinstance(item,basestring):
				if charMode:
					text=' '.join([x for x in unicode(item)])
				else:
					text=unicode(item)
				self.speakNow(text,lastIndex)
			elif isinstance(item,speech.IndexCommand):
				lastIndex=item.index
			elif isinstance(item,speech.CharacterModeCommand):
				charMode=item.state
			elif isinstance(item,speech.SpeechCommand):
				log.debugWarning("Unsupported speech command: %s"%item)
			else:
				log.error("Unknown speech: %s"%item)
